# Inspired by: https://docs.couchbase.com/python-sdk/current/howtos/error-handling.html
from enum import Enum
from functools import wraps
from random import randint
from time import sleep
from typing import Callable, Optional, Tuple


class RetryPolicy(Enum):
    FLAT = 1
    LINEAR = 2
    EXPONENTIAL = 3
    RANDOM = 4


def retry(
    attempts: int = 3,
    delay: float = 0.1,
    policy: RetryPolicy = RetryPolicy.FLAT,
    exceptions: Optional[Tuple[Exception]] = None,
) -> Callable:
    def handler(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for attempts_left in reversed(range(attempts)):
                try:
                    return func(*args, **kwargs)
                except Exception as _exc:
                    if exceptions is None or not isinstance(_exc, exceptions):
                        raise

                    if attempts_left == 0:
                        raise

                    backoff = delay
                    if policy == RetryPolicy.LINEAR:
                        backoff = delay * (attempts - attempts_left)
                    if policy == RetryPolicy.EXPONENTIAL:
                        backoff = (delay / 2) * (2 ** (attempts - attempts_left))
                    if policy == RetryPolicy.RANDOM:
                        backoff = delay * (randint(17, 233) / 100)

                    sleep(backoff)

        return wrapper

    return handler



from datetime import timedelta
import logging
from typing import Optional, Tuple, Union

from couchbase.auth import PasswordAuthenticator
from couchbase.bucket import Bucket
from couchbase.cluster import Cluster
from couchbase.collection import Collection
from couchbase.diagnostics import PingState
from couchbase.exceptions import (
    BucketAlreadyExistsException,
    CollectionAlreadyExistsException,
    ScopeAlreadyExistsException,
)
from couchbase.management.collections import CollectionSpec
from couchbase.management.logic.buckets_logic import CreateBucketSettings
from couchbase.options import (
    ClusterOptions,
    ClusterTimeoutOptions,
)
from couchbase.scope import Scope

from .exceptions import BucketNotSet, ClusterNotSet, ScopeNotSet
from .protocols import SessionProt
from .timeout import Timeout


class Session(SessionProt):
    """Create a Couchbase cluster session

    The session instance exposes methods for connecting, choosing bucket, scope, and collection for Couchbase operations

    Args:
        hostname (str):
            The hostname (or IP address) to connect to. Either should be without protocol,
            e.g. "localhost" or "127.0.0.1".
        username (str):
            The username to be used for authentication.
        password (str):
            The password to be used for authentication.

    Optional args:
        bucket (str):
            The bucket name to connect to.
        scope (str):
            The scope name to connect to. Defaults to "_default".
        collection (str):
            The collection to connect to. Defaults to `"_default".
        tls (bool):
            Whether connection should use TLS. Defaults to `false`
        timeout (:class:`~couchbase_helper.timeout.Timeout`):
            The timeout settings for this session instance. Defaults to `Timeout()`
        wan (bool):
            Whether connection profile "wan_development" should be applied.
        logger (:class:`logging.Logger`):
            Any logging instance to be used. Defaults to root logger.
    """

    def __init__(
        self,
        hostname: str,
        username: str,
        password: str,
        *,
        bucket: Optional[str] = None,
        scope: str = "_default",
        collection: str = "_default",
        tls: bool = False,
        timeout: Optional[Union[Timeout, Tuple[int, int, int], int]] = None,
        wan: bool = False,
        logger: Optional[logging.Logger] = None,
    ):
        # Initiate logger
        if logger is None:
            logger = logging.getLogger()
        self.logger = logger

        # Set default cluster values
        self._hostname = hostname
        self._username = username
        self._password = password
        self._cluster = None
        self._bucket = None
        self._bucket_name = bucket
        self._scope = None
        self._scope_name = scope
        self._collection = None
        self._collection_name = collection
        self._connected = False

        # Set the session's timeouts
        if isinstance(timeout, Timeout):
            self._timeout = timeout
        elif isinstance(timeout, tuple):
            self._timeout = Timeout(*timeout)
        elif isinstance(timeout, int):
            self._timeout = Timeout(timeout, timeout, timeout)
        elif timeout is None:
            self._timeout = Timeout()

        # Set connection options values
        self._tls = tls
        timeout_options = ClusterTimeoutOptions(
            connect_timeout=timedelta(seconds=self._timeout.connection)
        )
        self.options = ClusterOptions(
            authenticator=PasswordAuthenticator(
                username=self._username,
                password=self._password,
            ),
            enable_tls=tls,
            timeout_options=timeout_options,
            enable_tracing=True,
            show_queries=True,
        )
        if wan:
            self.options.apply_profile("wan_development")

    @property
    def connection_string(self) -> str:
        """Generates the cluster's connection string

        Returns:
            str: A combination of the protocol and hostname for the connection
        """
        return f"couchbase{'s' if self._tls else ''}://{self._hostname}"

    def connect(self):
        """Establish a connection to the cluster using the set bucket, scope, and collection."""
        self.logger.debug("- Connecting to cluster: %s", self.connection_string)

        if self._cluster is None:
            self._cluster = Cluster(self.connection_string, self.options)

        if self._bucket_name is not None:
            self.bucket = self._bucket_name

        if self._scope_name is not None:
            self.scope = self._scope_name

        if self._collection_name is not None:
            self.collection = self._collection_name

        self._cluster.wait_until_ready(timedelta(seconds=self._timeout.connection))
        self._connected = True

    def disconnect(self):
        """Shuts down the cluster instance and unsets class internal variables.

        According to Couchbase documentation...: "Use of this method is almost *always* unnecessary."
        It's available nonetheless.
        """
        if self._cluster is not None:
            self._cluster.close()
        self._cluster = None
        self._connected = False

    @property
    def connected(self) -> bool:
        """
        Returns:
            bool: Whether connected to cluster or not.
        """
        if self._cluster is None:
            return False

        return self._connected and self._cluster.connected

    @property
    def cluster(self) -> Cluster:
        """Returns the cluster instance"""
        return self._cluster

    @cluster.setter
    def cluster(self, value):
        connect = False
        if self._connected:
            connect = True
            self.disconnect()

        self._cluster = value

        if connect:
            self.connect()

    @property
    def bucket(self) -> Bucket:
        """Returns the bucket instance"""
        return self._bucket

    @bucket.setter
    def bucket(self, value):
        """Set the bucket instance"""
        self._bucket = self._cluster.bucket(value)
        self._bucket_name = value

    @property
    def bucket_name(self) -> str:
        """Returns the bucket name"""
        if self._bucket is None:
            raise BucketNotSet("no bucket set")

        return self._bucket_name

    def create_bucket(self, name, settings: CreateBucketSettings):
        """Create a bucket

        Args:
            name (str):
                The name of the bucket to create.
            settings (:class:`couchbase.management.logic.buckets_logic.BucketSettings`):
                The settings of the bucket to create.
        """
        if self._cluster is None:
            raise ClusterNotSet("no cluster set")

        bucket_manager = self._cluster.buckets()
        try:
            if settings.name is None:
                settings = CreateBucketSettings(name=name, **settings)

            bucket_manager.create_bucket(settings)
        except BucketAlreadyExistsException:
            pass

    @property
    def scope(self) -> Scope:
        """Returns the scope instance"""
        return self._scope

    @scope.setter
    def scope(self, value):
        """Set the scope instance"""
        if self._bucket is None:
            raise BucketNotSet("no bucket set")

        self._scope = self._bucket.scope(value)
        self._scope_name = value

    def create_scope(self, name):
        """Create a scope

        Args:
            name (str):
                The name of the scope to create.
        """
        if name == "_default":
            return

        collection_manager = self.bucket.collections()
        try:
            collection_manager.create_scope(name)
        except ScopeAlreadyExistsException:
            pass

    @property
    def collection(self) -> Collection:
        """Returns the collection instance"""
        return self._collection

    @collection.setter
    def collection(self, value):
        """Set the collection instance"""
        if self._scope is None:
            raise ScopeNotSet("no scope set")

        self._collection = self._scope.collection(value)
        self._collection_name = value

    def create_collection(self, name):
        """Create a collection

        Args:
            name (str):
                The name of the collection to create
        """
        if name == "_default":
            return

        if self._bucket is None:
            raise BucketNotSet("no bucket set")

        collection_manager = self._bucket.collections()
        collection_spec = CollectionSpec(self._collection_name, scope_name=name)
        try:
            collection_manager.create_collection(collection_spec)
        except CollectionAlreadyExistsException:
            pass

    def default_collection(self):
        """Use the default scope and collection"""
        self._scope = self.bucket.default_scope()
        self._scope_name = self._scope.name
        self._collection = self.bucket.default_collection()
        self._collection_name = self._collection.name

    def ping(self):
        """Ping the bucket to check for connection"""
        if self._bucket is None:
            raise BucketNotSet("no bucket set")

        result = self._bucket.ping()
        for endpoint, reports in result.endpoints.items():
            for report in reports:
                if not report.state == PingState.OK:
                    return False

        return True

    @property
    def timeout(self) -> Timeout:
        """Returns the timeout configuration

        Returns:
            Timeout:
                Will return the entire :class:`~couchbase_helper.timeout.Timeout` instance.
        """
        return self._timeout



import json
import logging
from typing import Protocol, runtime_checkable

import pytest
from couchbase_helper.protocols import SessionProt
from couchbase_helper import Timeout

from .couchbase_session import cb_instance, session
from .helpers import get_item, get_items


# Hack for checking if session complies
@runtime_checkable
class SessionProtExt(SessionProt, Protocol):
    pass


##################################################
# Test initiation and connection operations      #
##################################################
@pytest.mark.order(1)
def test_init():
    # Check session implements SessionProt
    assert isinstance(session, SessionProtExt)

    # Check session creates a logger instance
    assert isinstance(session.logger, logging.Logger)

    # Check session set's a default timeout
    _timeout = Timeout()
    assert isinstance(session.timeout, Timeout)
    assert session.timeout.connection == _timeout.connection
    assert session.timeout.kv == _timeout.kv
    assert session.timeout.query == _timeout.query

    # Check no connection was attempted/created
    assert not session.connected


@pytest.mark.order(2)
def test_connect():
    session.connect()
    assert session.connected


##################################################
# Test inserts and update operations             #
##################################################
def test_insert():
    key, document = get_item()
    assert cb_instance().insert(key, document)


def test_insert_duplicate():
    key, document = get_item()
    assert not cb_instance().insert(key, document)


def test_insert_multi():
    documents_raw = get_items(100)
    documents = {key: document for key, document in documents_raw}
    assert cb_instance().insert_multi(documents)


def test_upsert_existing():
    key, document = get_item()
    assert cb_instance().upsert(key, document)


def test_upsert():
    key, document = get_item("second")
    assert cb_instance().upsert(key, document)


def test_upsert_multi():
    documents_raw = get_items(100)
    documents = {key: document for key, document in documents_raw}
    assert cb_instance().upsert_multi(documents)


def test_replace():
    key, document = get_item("second")
    document["replaced"] = True
    assert cb_instance().replace(key, document)
    new_document = cb_instance().get(key)
    assert new_document.get("replaced", False)


def test_replace_multi():
    documents = {}
    for key, document in get_items(100):
        documents[key] = document
        documents[key]["replaced"] = True
    assert cb_instance().replace_multi(documents)
    new_documents = cb_instance().get_multi(list(documents.keys()))
    for new_document in new_documents:
        assert new_document.get("replaced", False)


@pytest.mark.order(after="test_insert")
def test_update():
    key, document = get_item()
    document["update"] = True
    assert cb_instance().upsert(key, document)


##################################################
# Test get and get_multi operations              #
##################################################
def test_get():
    key, document = get_item("get_test")
    cb_instance().insert(key, document)
    document_fetched = cb_instance().get(key)

    assert json.dumps(document) == json.dumps(document_fetched)
    assert cb_instance().remove(key)


def test_get_not_found():
    key, _ = get_item("get_not_found_test")
    assert not cb_instance().get(key)


def test_get_multi():
    documents_raw = get_items(100, "get_multi_test")
    documents = {key: document for key, document in documents_raw}
    cb_instance().insert_multi(documents)

    keys = [key for key, _ in documents_raw]
    documents_fetched = cb_instance().get_multi(keys, raw=True)
    assert len(documents) == len(documents_fetched)
    assert all(doc.key in keys for doc in documents_fetched)
    assert cb_instance().remove_multi(keys)


##################################################
# Test remove and remove_multi operations        #
##################################################
@pytest.mark.order(after="test_insert")
def test_remove():
    key, document = get_item()
    assert cb_instance().remove(key, document)


@pytest.mark.order(after="test_upsert")
def test_remove_second():
    key, document = get_item("second")
    assert cb_instance().remove(key, document)


@pytest.mark.order(after="test_remove")
def test_remove_not_found():
    key, document = get_item()
    assert not cb_instance().remove(key, document)


@pytest.mark.order(after="test_insert_multi")
def test_remove_multi():
    keys = [key for key, _ in get_items(100)]
    assert cb_instance().remove_multi(keys)


@pytest.mark.order(after="test_remove_multi")
def test_remove_multi_not_found():
    keys = [key for key, _ in get_items(100)]
    assert not cb_instance().remove_multi(keys)



import logging
import re
from typing import Any, Dict, List, Optional, Tuple, Union

from couchbase.n1ql import N1QLQuery, QueryScanConsistency
from couchbase.options import QueryOptions
from couchbase.result import QueryResult

from .options import build_opts
from .session import Session
from .timeout import Timeout

_TIMEOUT = Timeout()


class N1ql:
    """A Couchbase SQL++ (N1QL) class making for easier and safer query building

    Args:
        session (implements :class:`~couchbase_helper.protocols.SessionProt`)
            The cluster connection session
        logger (:class:`logging.logger`):
            The logging instance to use for log message. Defaults to the root logger.

    Usage:
        The class is intended used as chained methods, example:
        ```
        from couchbase_helper import Session
        from couchbase_helper.n1ql import N1ql

        session = Session("localhost", "user", "password", bucket="test")
        n1ql = N1ql(session)

        # select all restaurants from the "businesses" bucket
        restaurants = n1ql.select("name").from_("businesses").where("type=", "restaurant").get()
        for restaurant in restaurants:
            ...

        # select all butcher shops and bakeries from the "businesses" bucket
        shops = n1ql.select("name").from_("businesses").where("type=", "butcher").or_where("type=", "bakery").get()
        for shop in shops:
            ...
        ```
    """

    def __init__(self, session: Session, logger: Optional[logging.Logger] = None):
        if logger is None:
            logger = logging.getLogger()
        self.logger = logger

        self.session = session

        if not self.session.connected:
            self.session.connect()

        # Query runtime variables
        self._columns = ["*"]
        self._conditions: List[Tuple[str, str, Any]] = []
        self._distinct: bool = False
        self._limit: Optional[int] = None
        self._offset: Optional[int] = None
        self._session_reset: Dict[str, str] = {}

    def select(self, columns: Optional[Union[str, List[str]]] = None):
        """
        Set the column or columns to select

        Args:
            columns (Union[str, List[str]]):
                The column or columns to select in the query as a list of strings or comma separated string
        Returns:
            `self`
        """
        if columns is None:
            columns = ["*"]
        if isinstance(columns, str):
            columns = [col.strip() for col in columns.split(",")]

        self._columns = columns

        return self

    def distinct(self, enable: bool = True):
        """
        Will set query to select distinct values

        Args:
            enable (bool):
                Whether the distinct flag should be enabled (True) or not (False).
                This will always be reset to False when invoking the `get` method

        Returns:
            `self`
        """
        self._distinct = enable
        return self

    def from_(
        self,
        bucket: Optional[str] = None,
        scope: Optional[str] = None,
        collection: Optional[str] = None,
    ):
        if bucket is not None and bucket != self.session.bucket.name:
            self._session_reset["bucket"] = self.session.bucket.name
            self.session.bucket = bucket

        if scope is not None and scope != self.session.scope.name:
            self._session_reset["scope"] = self.session.scope.name
            self.session.scope = scope

        if collection is not None and collection != self.session.collection.name:
            self._session_reset["collection"] = self.session.collection.name
            self.session.collection = collection

        return self

    def where(self, key: str, value: Any):
        """
        Add an inclusive/and where condition

        Args:
            key (str):
                The key (and optionally operator) to add to the condition expression.
            value (Any):
                The value to compare with.

        Returns:
             `self`
        """
        self._where("AND", key, value)

        return self

    def or_where(self, key: str, value: Any):
        """
        Add an exclusive/or where condition

        Args:
            key (str):
                The key (and optionally operator) to add to the condition expression.
            value (Any):
                The value to compare with.

        Returns:
             `self`
        """
        self._where("OR", key, value)

        return self

    def orwhere(self, key: str, value: Any):
        """an alias for :class:`~.or_where`"""
        return self.or_where(key, value)

    # TODO ... : def not_(self, ...):

    def _where(self, type_: str, key: str, value: Any):
        """adds and/or where conditions"""
        cleaned_key = self._clean_key(key)
        column = self._enclose_reserved_word(cleaned_key)
        operator = None
        if value is None and not self._has_operator(key):
            operator = " IS NULL"
        if not self._has_operator(key) or not operator:
            operator = " ="

        processed_key = f"{column}{operator}"

        self._conditions.append((type_, processed_key, value))

    def limit(self, limit: int):
        """
        Add a limit to the query

        Args:
             limit (int):
                The limit of items a query can return

        Returns:
            `self`
        """
        try:
            self._limit = int(limit)
        except ValueError:
            pass

        return self

    def offset(self, offset: int):
        """
        Add an offset to the query

        Args:
             offset (int):
                The number of records the query must skip

        Returns:
            `self`
        """
        try:
            self._offset = int(offset)
        except ValueError:
            pass

        return self

    def skip(self, skip: int):
        """an alias for :class:`offset`"""
        return self.offset(skip)

    def rows(self, opts: Optional[QueryOptions] = None) -> Optional[QueryResult]:
        """
        Execute a select statement and return the fetched rows

        Args:
             opts (optional `couchbase.options.QueryOptions`):
                Any query options to use when executing the query.

        Returns:
            :class:`couchbase.result.QueryResult` iterable
        """
        # generate 'FROM' expression
        from_ = self._enclose_reserved_word(self.session.bucket.name)
        if (
            self.session.scope.name != "_default"
            or self.session.collection.name != "_default"
        ):
            from_ += f".{self._enclose_reserved_word(self.session.scope.name)}.{self._enclose_reserved_word(self.session.collection.name)}"

        ident = self.session.bucket.name[0:1]
        prefix = f"{ident}."

        # generate columns for selection
        columns = ", ".join(
            [
                f"{prefix if col != '*' else ''}{self._enclose_reserved_word(col)}"
                for col in self._columns
            ]
        )

        # generate where expression
        positional_arguments = []
        where = ""
        if self._conditions:
            for idx, condition in enumerate(self._conditions, 1):
                type_, key, value = condition
                positional_arguments.append(value)
                _where_part = f"{type_.upper() if len(where) else ''} {prefix}{key} ${idx}".strip()
                where += f"{_where_part} "
            where = f"WHERE {where}"

        # append limit
        limit = ""
        if self._limit is not None:
            limit = f"LIMIT {self._limit}"

        # append skip
        offset = ""
        if self._offset is not None:
            offset = f"OFFSET {self._offset}"

        # init QueryOptions
        if opts is None:
            opts = {}
        if positional_arguments:
            opts["positional_parameters"] = positional_arguments

        # generate the prepared statement
        statement = f"SELECT{' DISTINCT' if self._distinct else ''} {columns} FROM {from_} {ident} {where} {limit} {offset}".strip()

        # execute and return rows
        rows = None
        try:
            rows = self._execute(statement, opts).rows()
        except AttributeError:
            pass
        finally:
            self._reset()

        return rows

    def _execute(
        self,
        statement: str,
        opts: Optional[QueryOptions] = None,
        *,
        consistency: Optional[QueryScanConsistency] = None,
    ):
        """execute a statement"""
        # init QueryOptions
        if opts is None:
            opts = {}
        options = build_opts("query", opts=opts)

        try:
            # Initiate the N1QLQuery instance
            query = N1QLQuery(statement)
            if consistency is not None:
                query.consistency = consistency
            query.timeout = self.session.timeout.query
            return self.session.cluster.query(query.statement, **options)
        except Exception as _err:  # intentionally broad initially to see what exceptions are actually raised
            print(f"SQL++ exception happened ({type(_err).__name__}): {_err}")

        return None

    @staticmethod
    def _has_operator(string: str):
        """determine if a column/key has an operator"""
        return len(re.findall(r"(\s|<|>|!|=|is null|is not null)", string)) > 0

    @staticmethod
    def _clean_key(string):
        """clean a column/key from any non-word chars"""
        return re.sub(r"\W", "", string)

    def _reset(self):
        """resets class variables and session"""
        self._columns = ["*"]
        self._conditions = []
        self._distinct = False
        self._limit = None
        self._offset = None
        if self._session_reset:
            for prop, original in self._session_reset.items():
                setattr(self.session, prop, original)
            self._session_reset = {}

    @staticmethod
    def _enclose_reserved_word(word):
        # TODO: find a better place for this list ...
        """enclose reserved words"""
        reserved = (
            "_INDEX_CONDITION",
            "_INDEX_KEY",
            "ADVISE",
            "ALL",
            "ALTER",
            "ANALYZE",
            "AND",
            "ANY",
            "ARRAY",
            "AS",
            "ASC",
            "AT",
            "BEGIN",
            "BETWEEN",
            "BINARY",
            "BOOLEAN",
            "BREAK",
            "BUCKET",
            "BUILD",
            "BY",
            "CACHE",
            "CALL",
            "CASE",
            "CAST",
            "CLUSTER",
            "COLLATE",
            "COLLECTION",
            "COMMIT",
            "COMMITTED",
            "CONNECT",
            "CONTINUE",
            "CORRELATED",
            "COVER",
            "CREATE",
            "CURRENT",
            "CYCLE",
            "DATABASE",
            "DATASET",
            "DATASTORE",
            "DECLARE",
            "DECREMENT",
            "DEFAULT",
            "DELETE",
            "DERIVED",
            "DESC",
            "DESCRIBE",
            "DISTINCT",
            "DO",
            "DROP",
            "EACH",
            "ELEMENT",
            "ELSE",
            "END",
            "ESCAPE",
            "EVERY",
            "EXCEPT",
            "EXCLUDE",
            "EXECUTE",
            "EXISTS",
            "EXPLAIN",
            "FALSE",
            "FETCH",
            "FILTER",
            "FIRST",
            "FLATTEN",
            "FLATTEN_KEYS",
            "FLUSH",
            "FOLLOWING",
            "FOR",
            "FORCE",
            "FROM",
            "FTS",
            "FUNCTION",
            "GOLANG",
            "GRANT",
            "GROUP",
            "GROUPS",
            "GSI",
            "HASH",
            "HAVING",
            "IF",
            "IGNORE",
            "ILIKE",
            "IN",
            "INCLUDE",
            "INCREMENT",
            "INDEX",
            "INFER",
            "INLINE",
            "INNER",
            "INSERT",
            "INTERSECT",
            "INTO",
            "IS",
            "ISOLATION",
            "JAVASCRIPT",
            "JOIN",
            "KEY",
            "KEYS",
            "KEYSPACE",
            "KNOWN",
            "LANGUAGE",
            "LAST",
            "LATERAL",
            "LEFT",
            "LET",
            "LETTING",
            "LEVEL",
            "LIKE",
            "LIMIT",
            "LSM",
            "MAP",
            "MAPPING",
            "MATCHED",
            "MATERIALIZED",
            "MAXVALUE",
            "MERGE",
            "MINVALUE",
            "MISSING",
            "NAMESPACE",
            "NEST",
            "NEXT",
            "NEXTVAL",
            "NL",
            "NO",
            "NOT",
            "NTH_VALUE",
            "NULL",
            "NULLS",
            "NUMBER",
            "OBJECT",
            "OFFSET",
            "ON",
            "OPTION",
            "OPTIONS",
            "OR",
            "ORDER",
            "OTHERS",
            "OUTER",
            "OVER",
            "PARSE",
            "PARTITION",
            "PASSWORD",
            "PATH",
            "POOL",
            "PRECEDING",
            "PREPARE",
            "PREV",
            "PREVIOUS",
            "PREVVAL",
            "PRIMARY",
            "PRIVATE",
            "PRIVILEGE",
            "PROBE",
            "PROCEDURE",
            "PUBLIC",
            "RANGE",
            "RAW",
            "READ",
            "REALM",
            "RECURSIVE",
            "REDUCE",
            "RENAME",
            "REPLACE",
            "RESPECT",
            "RESTART",
            "RESTRICT",
            "RETURN",
            "RETURNING",
            "REVOKE",
            "RIGHT",
            "ROLE",
            "ROLLBACK",
            "ROW",
            "ROWS",
            "SATISFIES",
            "SAVEPOINT",
            "SCHEMA",
            "SCOPE",
            "SELECT",
            "SELF",
            "SEQUENCE",
            "SET",
            "SHOW",
            "SOME",
            "START",
            "STATISTICS",
            "STRING",
            "SYSTEM",
            "THEN",
            "TIES",
            "TO",
            "TRAN",
            "TRANSACTION",
            "TRIGGER",
            "TRUE",
            "TRUNCATE",
            "UNBOUNDED",
            "UNDER",
            "UNION",
            "UNIQUE",
            "UNKNOWN",
            "UNNEST",
            "UNSET",
            "UPDATE",
            "UPSERT",
            "USE",
            "USER",
            "USERS",
            "USING",
            "VALIDATE",
            "VALUE",
            "VALUED",
            "VALUES",
            "VECTOR",
            "VIA",
            "VIEW",
            "WHEN",
            "WHERE",
            "WHILE",
            "WINDOW",
            "WITH",
            "WITHIN",
            "WORK",
            "XOR",
        )

        if word.upper() in reserved:
            word = f"`{word}`"

        if "." in word:
            word = f"`{word}`"

        return word



from couchbase_helper.n1ql import N1ql

from .couchbase_session import n1ql_instance


##################################################
# Test instance                                  #
##################################################
def test_instance():
    assert isinstance(n1ql_instance(), N1ql)


##################################################
# Test select                                    #
##################################################
def test_select_default():
    assert n1ql_instance()._columns == ["*"]


def test_select_custom_string():
    n1ql_instance().select("kings,queens")
    assert n1ql_instance()._columns == ["kings", "queens"]


def test_select_custom_list():
    n1ql_instance().select(["horses", "donkeys"])
    assert n1ql_instance()._columns == ["horses", "donkeys"]


##################################################
# Test distinct                                  #
##################################################
def test_distinct():
    n1ql_instance().distinct()
    assert n1ql_instance()._distinct


def test_not_distinct():
    n1ql_instance().distinct(False)
    assert not n1ql_instance()._distinct


##################################################
# Test from_                                     #
##################################################
def test_from_bucket():
    bucket_name = "bucketname"
    n1ql_instance().from_(bucket_name)
    assert n1ql_instance().session.bucket.name == bucket_name
    n1ql_instance()._reset()


def test_from_bucket_scope():
    bucket_name = "bucketname2"
    scope_name = "scope2"
    n1ql_instance().from_(bucket_name, scope_name)
    assert n1ql_instance().session.bucket.name == bucket_name
    assert n1ql_instance().session.scope.name == scope_name
    n1ql_instance()._reset()


def test_from_bucket_scope_collection():
    bucket_name = "bucketname3"
    scope_name = "scope3"
    collection_name = "collection3"
    n1ql_instance().from_(bucket_name, scope_name, collection_name)
    assert n1ql_instance().session.bucket.name == bucket_name
    assert n1ql_instance().session.scope.name == scope_name
    assert n1ql_instance().session.collection.name == collection_name


##################################################
# Test where                                     #
##################################################
def test_where():
    n1ql_instance().where("type=", "random")
    assert n1ql_instance()._conditions == [("AND", "type =", "random")]


def test_where_or():
    n1ql_instance().or_where("type=", "not-random")
    assert n1ql_instance()._conditions == [
        ("AND", "type =", "random"),
        ("OR", "type =", "not-random"),
    ]


##################################################
# Test limit                                     #
##################################################
def test_limit():
    n1ql_instance().limit("11")
    assert n1ql_instance()._limit == 11
    n1ql_instance().limit(10)
    assert n1ql_instance()._limit == 10
    n1ql_instance().limit("invalid-string")
    assert n1ql_instance()._limit == 10


##################################################
# Test offset/skip                               #
##################################################
def test_offset():
    n1ql_instance().offset(3)
    assert n1ql_instance()._offset == 3
    n1ql_instance().offset("4")
    assert n1ql_instance()._offset == 4
    n1ql_instance().offset("invalid-string")
    assert n1ql_instance()._offset == 4


##################################################
# Test rows                                      #
##################################################
def test_rows():
    assert n1ql_instance().rows() is None


##################################################
# Test reset                                     #
##################################################
def test_reset():
    assert n1ql_instance()._columns == ["*"]
    assert not n1ql_instance()._distinct
    assert n1ql_instance().session.bucket.name == "test"
    assert n1ql_instance().session.scope.name == "_default"
    assert n1ql_instance().session.collection.name == "_default"
    assert n1ql_instance()._conditions == []
    assert n1ql_instance()._limit is None
    assert n1ql_instance()._offset is None



from .helper import CouchbaseHelper
from .session import Session
from .timeout import Timeout

__all__ = ["CouchbaseHelper", "Session", "Timeout"]
__author__ = "Thomas 'sitzz' Vang <sitzzdk@gmail.com>"
__version__ = "0.1.2"



class Timeout:
    """A simple container for timeout values of different operations.

    Has some sane defaults.

    Args:
        connection (int):
            The timeout of connection operations.
        kv (int):
            The timeout used for key-value operations.
        query (int):
            The timeout used for query operations.
    """

    def __init__(self, connection: int = 10, kv: int = 5, query: int = 30):
        self._connection = connection
        self._kv = kv
        self._query = query

    @classmethod
    def init(cls, connection: int = 10, kv: int = 5, query: int = 30):
        """Statically initiate class instance"""
        return cls(connection, kv, query)

    @property
    def connection(self) -> int:
        """Returns the connection timeout value"""
        return self._connection

    @property
    def kv(self) -> int:
        """Returns the key-value timeout value"""
        return self._kv

    @property
    def query(self) -> int:
        """Returns the query timeout value"""
        return self._query



from typing import Protocol

from couchbase.bucket import Collection, Scope
from couchbase.cluster import Bucket, Cluster

from .timeout import Timeout


class SessionProt(Protocol):
    def connect(self): ...

    def disconnect(self): ...

    @property
    def connected(self): ...

    @property
    def cluster(self) -> Cluster: ...

    @property
    def bucket(self) -> Bucket: ...

    @property
    def bucket_name(self) -> str: ...

    @property
    def collection(self) -> Collection: ...

    @property
    def scope(self) -> Scope: ...

    def ping(self) -> bool: ...

    @property
    def timeout(self) -> Timeout: ...



from datetime import timedelta
from typing import Any, Dict, Optional, Union

from couchbase.options import (
    GetMultiOptions,
    GetOptions,
    InsertMultiOptions,
    InsertOptions,
    QueryOptions,
    RemoveMultiOptions,
    RemoveOptions,
    ReplaceMultiOptions,
    ReplaceOptions,
    UpsertMultiOptions,
    UpsertOptions,
    ViewOptions,
)

from .session import Session
from .timeout import Timeout

_TIMEOUT = Timeout()
_TYPES = {
    "get": GetOptions,
    "get_multi": GetMultiOptions,
    "insert": InsertOptions,
    "insert_multi": InsertMultiOptions,
    "query": QueryOptions,
    "remove": RemoveOptions,
    "remove_multi": RemoveMultiOptions,
    "replace": ReplaceOptions,
    "replace_multi": ReplaceMultiOptions,
    "upsert": UpsertOptions,
    "upsert_multi": UpsertMultiOptions,
    "view": ViewOptions,
}


def build_opts(
    type_: str,
    *,
    opts: Optional[Dict[str, Any]] = None,
    expiry: Optional[Union[int, timedelta]] = None,
    session: Optional[Session] = None,
) -> Union[
    InsertOptions,
    InsertMultiOptions,
    UpsertOptions,
    UpsertMultiOptions,
    ReplaceOptions,
    ReplaceMultiOptions,
    GetOptions,
    GetMultiOptions,
    QueryOptions,
    RemoveOptions,
    RemoveMultiOptions,
    ViewOptions,
]:
    """Generates operation options for specified operation type.

    Args:
        type_ (str):
            The type of operation to return options for.
        opts (Dict[str, Any]):
            Optional options to use for initiating the operation options instance.
        expiry (int | timedelta | None):
            Optional general document expiry to use for the operations.
        session (couchbase_helper.Session):
            Optional session to fetch timeout settings from

    Returns:
        Dict[str, Any]
    """
    if type_ not in _TYPES:
        raise AttributeError(f"unknown options type {type_}")

    base_options = _TYPES[type_]

    if opts is None:
        opts = {}

    if session is not None:
        timeout = session.timeout
    else:
        timeout = _TIMEOUT

    if "timeout" not in opts:
        if type_ in ("query", "view"):
            opts["timeout"] = timedelta(seconds=timeout.query)
        else:
            opts["timeout"] = timedelta(seconds=timeout.kv)

    if expiry is not None:
        if isinstance(expiry, int):
            expiry = timedelta(seconds=expiry)
        opts["expiry"] = expiry

    return base_options(**opts)



from datetime import timedelta
import logging
from typing import Any, Dict, List, Optional, Union

from couchbase.diagnostics import ServiceType
from couchbase.exceptions import DocumentExistsException, DocumentNotFoundException
from couchbase.options import (
    InsertOptions,
    UpsertOptions,
    WaitUntilReadyOptions,
)
from couchbase.result import GetResult, MultiGetResult

from ._types import JSONType
from .options import build_opts
from .protocols import SessionProt


class CouchbaseHelper:
    """A couchbase helper class to simplify document operations

    Args:
        session (implements :class:`couchbase_helper.protocols.SessionProt`):
            The cluster connection session
        logger (:class:`logging.logger`):
            The logging instance to use for log message. Defaults to the root logger.
    """

    def __init__(
        self,
        session: SessionProt,
        logger: Optional[logging.Logger] = None,
    ):
        if logger is None:
            logger = logging.getLogger()
        self.logger = logger

        self.session = session

        if not self.session.connected:
            self.session.connect()

    def insert(
        self,
        key: str,
        value: JSONType,
        expiry: Optional[Union[int, timedelta]] = None,
        opts: Optional[Dict[str, Any]] = None,
    ) -> bool:
        """Insert a single document. Will fail if document already exists.

        Args:
            key (str):
                The key of the document to save.
            value (`couchbase_helper._types.JSONType`):
                The value of the document to save.
            expiry (int | `:class:`datetime.timedelta`):
                The expiry of the document to save.
            opts (Dict[str, Any]):
                The operation options to use when saving document.

        Returns:
            (bool):
                The status of the insert operation.
        """
        args = {
            "key": key,
            "value": value,
            "opts": build_opts("insert", opts=opts, expiry=expiry),
        }

        try:
            self.session.cluster.wait_until_ready(
                timedelta(self.session.timeout.kv),
                WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]),
            )
            self.session.collection.insert(**args)
            return True
        except DocumentExistsException:
            return False

    def insert_multi(
        self,
        documents: Dict[str, JSONType],
        expiry: Optional[Union[int, timedelta]] = None,
        opts: Optional[Dict[str, Any]] = None,
        per_key_opts: Optional[Dict[str, InsertOptions]] = None,
    ):
        """Insert multiple documents, for each key-value pair in the `documents` dictionary
        a document will be created.

        Args:
            documents (Dict[str, JSONType]):
                A dictionary of the documents to be saved.
            expiry (int | `:class:`datetime.timedelta`):
                The expiry of the documents to save.
            opts (Dict[str, Any]):
                The operation options to use when saving document.
            per_key_opts (Dict[str, :class:`couchbase.options.InsertOptions`]):
                A dictionary of :class:`couchbase.options.InsertOptions` per document key.

        Returns:
            (bool):
                The status of the insert operations. Will return `True` if all operations
                were successful, `False` otherwise.
        """
        if opts is None:
            opts = {}

        if per_key_opts is not None:
            for key, val in per_key_opts.items():
                per_key_opts[key] = build_opts("insert_multi", opts=val)

            opts["per_key_options"] = per_key_opts

        args = {
            "keys_and_docs": documents,
            "opts": build_opts("insert_multi", opts=opts, expiry=expiry),
        }
        try:
            self.session.cluster.wait_until_ready(
                timedelta(self.session.timeout.kv),
                WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]),
            )
            result = self.session.collection.insert_multi(**args)
            if result.all_ok:
                return True

            for key, exception in result.exceptions.items():
                self.logger.error("unable to add document %s: %s", key, exception)
        except Exception as _err:
            self.logger.error(
                "unhandled exception (%s): %s", type(_err).__name__, _err.args[0]
            )

        return False

    def upsert(
        self,
        key: str,
        value: JSONType,
        expiry: Optional[Union[int, timedelta]] = None,
        opts: Optional[Dict[str, Any]] = None,
    ):
        """Update or insert a single document.

        Args:
            key (str):
                The key of the document to save.
            value (`couchbase_helper._types.JSONType`):
                The value of the document to save.
            expiry (int | `:class:`datetime.timedelta`):
                The expiry of the document to save.
            opts (Dict[str, Any]):
                The operation options to use when saving document.

        Returns:
            (bool):
                The status of the insert operation.
        """
        args = {
            "key": key,
            "value": value,
            "opts": build_opts("upsert", opts=opts, expiry=expiry),
        }

        try:
            self.session.cluster.wait_until_ready(
                timedelta(self.session.timeout.kv),
                WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]),
            )
            self.session.collection.upsert(**args)
            return True
        except DocumentNotFoundException:
            return False

    def upsert_multi(
        self,
        documents: Dict[str, JSONType],
        expiry: Optional[Union[int, timedelta]] = None,
        opts: Optional[Dict[str, Any]] = None,
        per_key_opts: Optional[Dict[str, UpsertOptions]] = None,
    ) -> bool:
        """Update or insert multiple documents, for each key-value pair in the
        `documents` dictionary a document will be updated or created.

        Args:
            documents (Dict[str, JSONType]):
                A dictionary of the documents to be saved.
            expiry (int | `:class:`datetime.timedelta`):
                The expiry of the documents to save.
            opts (Dict[str, Any]):
                The operation options to use when saving document.
            per_key_opts (Dict[str, :class:`couchbase.options.UpsertOptions`]):
                A dictionary of :class:`couchbase.options.UpsertOptions` per document key.

        Returns:
            (bool):
                The status of the upsert operations. Will return `True` if all operations
                were successful, `False` otherwise.
        """
        if opts is None:
            opts = {}

        if per_key_opts is not None:
            for key, val in per_key_opts.items():
                per_key_opts[key] = build_opts("upsert", opts=val)

            opts["per_key_options"] = per_key_opts

        args = {
            "keys_and_docs": documents,
            "opts": build_opts("upsert_multi", opts=opts, expiry=expiry),
        }
        try:
            self.session.cluster.wait_until_ready(
                timedelta(self.session.timeout.kv),
                WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]),
            )
            result = self.session.collection.upsert_multi(**args)
            if result.all_ok:
                return True

            for key, exception in result.exceptions.items():
                self.logger.error("unable to add document %s: %s", key, exception)
        except Exception as _err:
            self.logger.error(
                "unhandled exception (%s): %s", type(_err).__name__, _err.args[0]
            )

        return False

    def replace(
        self,
        key: str,
        value: JSONType,
        expiry: Optional[Union[int, timedelta]] = None,
        opts: Optional[Dict[str, Any]] = None,
    ):
        """Replace a single document.

        Args:
            key (str):
                The key of the document to save.
            value (`couchbase_helper._types.JSONType`):
                The value of the document to save.
            expiry (int | `:class:`datetime.timedelta`):
                The expiry of the document to save.
            opts (Dict[str, Any]):
                The operation options to use when saving document.

        Returns:
            (bool):
                The status of the insert operation.
        """
        args = {
            "key": key,
            "value": value,
            "opts": build_opts("replace", opts=opts, expiry=expiry),
        }

        try:
            self.session.cluster.wait_until_ready(
                timedelta(self.session.timeout.kv),
                WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]),
            )
            self.session.collection.upsert(**args)
            return True
        except DocumentNotFoundException:
            return False

    def replace_multi(
        self,
        documents: Dict[str, JSONType],
        expiry: Optional[Union[int, timedelta]] = None,
        opts: Optional[Dict[str, Any]] = None,
        per_key_opts: Optional[Dict[str, UpsertOptions]] = None,
    ) -> bool:
        """Update or insert multiple documents, for each key-value pair in the
        `documents` dictionary a document will be updated or created.

        Args:
            documents (Dict[str, JSONType]):
                A dictionary of the documents to be saved.
            expiry (int | `:class:`datetime.timedelta`):
                The expiry of the documents to save.
            opts (Dict[str, Any]):
                The operation options to use when saving document.
            per_key_opts (Dict[str, :class:`couchbase.options.UpsertOptions`]):
                A dictionary of :class:`couchbase.options.UpsertOptions` per document key.

        Returns:
            (bool):
                The status of the upsert operations. Will return `True` if all operations
                were successful, `False` otherwise.
        """
        if opts is None:
            opts = {}

        if per_key_opts is not None:
            for key, val in per_key_opts.items():
                per_key_opts[key] = build_opts("upsert", opts=val)

            opts["per_key_options"] = per_key_opts

        args = {
            "keys_and_docs": documents,
            "opts": build_opts("replace_multi", opts=opts, expiry=expiry),
        }
        try:
            self.session.cluster.wait_until_ready(
                timedelta(self.session.timeout.kv),
                WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]),
            )
            result = self.session.collection.replace_multi(**args)
            if result.all_ok:
                return True

            for key, exception in result.exceptions.items():
                self.logger.error("unable to replace document %s: %s", key, exception)
        except Exception as _err:
            self.logger.error(
                "unhandled exception (%s): %s", type(_err).__name__, _err.args[0]
            )

        return False

    def get(
        self, key: str, opts: Optional[Dict[str, Any]] = None, *, raw: bool = False
    ) -> Optional[Union[GetResult, Dict[Any, Any]]]:
        """Get a single document

        Args:
            key (str):
                The key of the document to fetch.
            opts (Dict[str, Any]):
                The operation options to use when fetching the document.
            raw (bool):
                Whether to return the raw Couchbase response. Will return only the value
                as a dictionary otherwise.

        Returns:
            :class:`couchbase.result.GetResult` | Dict[Any, Any] | None
        """
        args = {"key": key, "opts": build_opts("get", opts=opts)}

        try:
            self.session.cluster.wait_until_ready(
                timedelta(self.session.timeout.kv),
                WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]),
            )
            document = self.session.collection.get(**args)
            return document if raw else document.content_as[dict]
        except DocumentNotFoundException:
            return None

    def get_multi(
        self,
        keys: List[str],
        opts: Optional[Dict[str, Any]] = None,
        *,
        raw: bool = False,
    ) -> Optional[Union[MultiGetResult, List[Dict[Any, Any]]]]:
        """Get multiple document

        Args:
            keys (List[str]):
                The key of the document to fetch.
            opts (Dict[str, Any]):
                The operation options to use when fetching the documents.
            raw (bool):
                Whether to return the raw Couchbase responses. Will return only the values
                as a dictionaries otherwise.

        Returns:
            :class:`couchbase.result.MultiGetResult` | List[Dict[Any, Any]] | None
        """
        args = {"keys": keys, "opts": build_opts("get_multi", opts=opts)}

        try:
            ret = []
            self.session.cluster.wait_until_ready(
                timedelta(self.session.timeout.kv),
                WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]),
            )
            documents = self.session.collection.get_multi(**args).results
            for _, document in documents.items():
                ret.append(document if raw else document.content_as[dict])

            return ret
        except DocumentNotFoundException:
            return None

    def remove(self, key: str, opts: Optional[Dict[str, Any]] = None) -> bool:
        """Remove a single key

        Args:
            key (str):
                The key of the document to remove.
            opts (Dict[str, Any]):
                The operation options to use when removing the document.

        Returns:
            (bool):
                The status of the remove operation.
        """
        args = {"key": key, "opts": build_opts("remove", opts=opts)}

        self.session.cluster.wait_until_ready(
            timedelta(self.session.timeout.kv),
            WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]),
        )
        try:
            self.session.collection.remove(**args)
            return True
        except DocumentNotFoundException:
            pass

        return False

    def remove_multi(
        self, keys: List[str], opts: Optional[Dict[str, Any]] = None
    ) -> bool:
        """Remove multiple keys

        Args:
            keys (List[str]):
                The keys of the documents to remove.
            opts (Dict[str, Any]):
                The operation options to use when removing the documents.

        Returns:
            (bool):
                The status of the remove operations.
        """
        args = {"keys": keys, "opts": build_opts("remove_multi", opts=opts)}

        self.session.cluster.wait_until_ready(
            timedelta(self.session.timeout.kv),
            WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]),
        )
        try:
            result = self.session.collection.remove_multi(**args)

            if result.all_ok:
                return True

            for key, exception in result.exceptions.items():
                self.logger.error("unable to remove document %s: %s", key, exception)
        except Exception as _err:
            self.logger.error(
                "unhandled exception (%s): %s", type(_err).__name__, _err.args[0]
            )

        return False

    def delete(self, *args, **kwargs):
        """Alias for :class:`~.remove`"""
        return self.remove(*args, **kwargs)

    def delete_multi(self, *args, **kwargs):
        """Alias for :class:`~.remove_multi`"""
        return self.remove_multi(*args, **kwargs)

    def view_query(
        self,
        design_doc: str,
        view_name: str,
        *,
        limit: Optional[int] = None,
        skip: Optional[int] = None,
        opts: Optional[Dict[str, Any]] = None,
    ) -> Optional[List[Dict[Any, Any]]]:
        if opts is None:
            opts = {}
        if limit is not None:
            opts["limit"] = limit
        if skip is not None:
            opts["skip"] = skip

        total_rows = 0
        try:
            self.session.cluster.wait_until_ready(
                timedelta(self.session.timeout.kv),
                WaitUntilReadyOptions(service_types=[ServiceType.View]),
            )
            query = self.session.bucket.view_query(
                design_doc=design_doc,
                view_name=view_name,
                **build_opts("view", opts=opts),
            )
            query_metadata = query.metadata()
            if query_metadata is not None:
                total_rows = query_metadata.total_rows()
            if query_metadata is None or (skip is not None and total_rows < skip):
                return query.rows()
        except Exception as _exc:
            self.logger.error(
                "CouchBase view query error (%s): %s",
                type(_exc).__name__,
                str(_exc.args[0]),
            )

        return None
